@@ -35,12 +35,17 @@
 #ifndef ARCH_H
 #define ARCH_H
 
+#include "speex/speex_types.h"
+
 #define ABS(x) ((x) < 0 ? (-(x)) : (x))
+#define ABS16(x) ((x) < 0 ? (-(x)) : (x))
+#define MAX16(a,b) ((a) > (b) ? (a) : (b))
+#define ABS32(x) ((x) < 0 ? (-(x)) : (x))
 
 #ifdef FIXED_POINT
 
-typedef signed short spx_word16_t;
-typedef signed int   spx_word32_t;
+typedef spx_int16_t spx_word16_t;
+typedef spx_int32_t   spx_word32_t;
 #ifdef _MSC_VER
 typedef __int64      spx_word64_t;
 #else
@@ -51,6 +56,8 @@ typedef spx_word16_t spx_coef_t;
 typedef spx_word16_t spx_lsp_t;
 typedef spx_word32_t spx_sig_t;
 
+#define Q15ONE 32767
+
 #define LPC_SCALING  8192
 #define SIG_SCALING  16384
 #define LSP_SCALING  8192.
@@ -59,21 +66,31 @@ typedef spx_word32_t spx_sig_t;
 #define GAIN_SCALING_1 0.015625
 
 #define LPC_SHIFT    13
+#define LSP_SHIFT    13
 #define SIG_SHIFT    14
 
 #define VERY_SMALL 0
+#define VERY_LARGE32 ((spx_word32_t)2147483647)
+#define VERY_LARGE16 ((spx_word16_t)32767)
 
 
+#ifdef FIXED_DEBUG
+#include "fixed_debug.h"
+#else
+
+#include "fixed_generic.h"
+
 #ifdef ARM5E_ASM
 #include "fixed_arm5e.h"
 #elif defined (ARM4_ASM)
 #include "fixed_arm4.h"
-#elif defined (FIXED_DEBUG)
-#include "fixed_debug.h"
-#else
-#include "fixed_generic.h"
+#elif defined (ARM5E_ASM)
+#include "fixed_arm5e.h"
+#elif defined (BFIN_ASM)
+#include "fixed_bfin.h"
 #endif
 
+#endif
 
 
 #else
@@ -86,17 +103,34 @@ typedef float spx_word16_t;
 typedef float spx_word32_t;
 typedef float spx_word64_t;
 
-#define LPC_SCALING  1.
-#define SIG_SCALING  1.
-#define LSP_SCALING  1.
-#define GAMMA_SCALING 1.
-#define GAIN_SCALING 1.
-#define GAIN_SCALING_1 1.
+#define Q15ONE 1.0f
+#define LPC_SCALING  1.f
+#define SIG_SCALING  1.f
+#define LSP_SCALING  1.f
+#define GAMMA_SCALING 1.f
+#define GAIN_SCALING 1.f
+#define GAIN_SCALING_1 1.f
 
 #define LPC_SHIFT    0
+#define LSP_SHIFT    0
 #define SIG_SHIFT    0
 
-#define VERY_SMALL 1e-30
+#define VERY_SMALL 1e-15f
+#define VERY_LARGE32 1e15f
+#define VERY_LARGE16 1e15f
+
+#define NEG16(x) (-(x))
+#define NEG32(x) (-(x))
+#define EXTRACT16(x) (x)
+#define EXTEND32(x) (x)
+#define SHR16(a,shift) (a)
+#define SHL16(a,shift) (a)
+#define SHR32(a,shift) (a)
+#define SHL32(a,shift) (a)
+#define PSHR16(a,shift) (a)
+#define PSHR32(a,shift) (a)
+#define SATURATE16(x,a) (x)
+#define SATURATE32(x,a) (x)
 
 #define PSHR(a,shift)       (a)
 #define SHR(a,shift)       (a)
@@ -109,8 +143,8 @@ typedef float spx_word64_t;
 #define SUB32(a,b) ((a)-(b))
 #define ADD64(a,b) ((a)+(b))
 #define MULT16_16_16(a,b)     ((a)*(b))
-#define MULT16_16(a,b)     ((a)*(b))
-#define MAC16_16(c,a,b)     ((c)+(a)*(b))
+#define MULT16_16(a,b)     ((spx_word32_t)(a)*(spx_word32_t)(b))
+#define MAC16_16(c,a,b)     ((c)+(spx_word32_t)(a)*(spx_word32_t)(b))
 
 #define MULT16_32_Q11(a,b)     ((a)*(b))
 #define MULT16_32_Q13(a,b)     ((a)*(b))
@@ -121,10 +155,12 @@ typedef float spx_word64_t;
 #define MAC16_32_Q15(c,a,b)     ((c)+(a)*(b))
 
 #define MAC16_16_Q11(c,a,b)     ((c)+(a)*(b))
-#define MULT16_16_Q11(a,b)     ((a)*(b))
+#define MAC16_16_Q13(c,a,b)     ((c)+(a)*(b))
+#define MULT16_16_Q11_32(a,b)     ((a)*(b))
 #define MULT16_16_Q13(a,b)     ((a)*(b))
 #define MULT16_16_Q14(a,b)     ((a)*(b))
 #define MULT16_16_Q15(a,b)     ((a)*(b))
+#define MULT16_16_P15(a,b)     ((a)*(b))
 
 #define DIV32_16(a,b)     ((a)/(b))
 #define DIV32(a,b)     ((a)/(b))
@@ -133,5 +169,19 @@ typedef float spx_word64_t;
 #endif
 
 
+#ifdef CONFIG_TI_C55X
+
+/* 2 on TI C5x DSP */
+#define BYTES_PER_CHAR 2 
+#define BITS_PER_CHAR 16
+#define LOG2_BITS_PER_CHAR 4
+
+#else 
+
+#define BYTES_PER_CHAR 1
+#define BITS_PER_CHAR 8
+#define LOG2_BITS_PER_CHAR 3
+
+#endif
 
 #endif
