@@ -36,7 +36,7 @@
 #endif
 
 #include <math.h>
-#include <speex/speex_preprocess.h>
+#include "speex/speex_preprocess.h"
 #include "misc.h"
 #include "smallft.h"
 
@@ -102,10 +102,11 @@ static float hypergeom_gain(float x)
    if (x>9.5)
       return 1+.12/x;
 
-   integer = floor(x);
-   frac = x-integer;
+   integer = floor(2*x);
+   frac = 2*x-integer;
    ind = (int)integer;
-   
+   /*if (ind > 20 || ind < 0)
+   fprintf (stderr, "error: %d %f\n", ind, x);*/
    return ((1-frac)*table[ind] + frac*table[ind+1])/sqrt(x+.0001f);
 }
 
@@ -269,7 +270,7 @@ void speex_preprocess_state_destroy(SpeexPreprocessState *st)
    speex_free(st);
 }
 
-static void update_noise(SpeexPreprocessState *st, float *ps, int *echo)
+static void update_noise(SpeexPreprocessState *st, float *ps, float *echo)
 {
    int i;
    float beta;
@@ -281,10 +282,10 @@ static void update_noise(SpeexPreprocessState *st, float *ps, int *echo)
    if (!echo)
    {
       for (i=0;i<st->ps_size;i++)
-         st->noise[i] = (1.f-beta)*st->noise[i] + beta*ps[i];   
+         st->noise[i] = (1.f-beta)*st->noise[i] + beta*ps[i];
    } else {
       for (i=0;i<st->ps_size;i++)
-         st->noise[i] = (1.f-beta)*st->noise[i] + beta*max(0.f,ps[i]-echo[i]); 
+         st->noise[i] = (1.f-beta)*st->noise[i] + beta*max(1.f,ps[i]-echo[i]); 
 #if 0
       for (i=0;i<st->ps_size;i++)
          st->noise[i] = 0;
@@ -547,7 +548,7 @@ static void speex_compute_agc(SpeexPreprocessState *st, float mean_prior)
    
 }
 
-static void preprocess_analysis(SpeexPreprocessState *st, short *x)
+static void preprocess_analysis(SpeexPreprocessState *st, spx_int16_t *x)
 {
    int i;
    int N = st->ps_size;
@@ -620,7 +621,7 @@ static void update_noise_prob(SpeexPreprocessState *st)
 
 #define NOISE_OVERCOMPENS 1.4
 
-int speex_preprocess(SpeexPreprocessState *st, short *x, int *echo)
+int speex_preprocess(SpeexPreprocessState *st, spx_int16_t *x, float *echo)
 {
    int i;
    int is_speech=1;
@@ -648,7 +649,7 @@ int speex_preprocess(SpeexPreprocessState *st, short *x, int *echo)
    /* Deal with residual echo if provided */
    if (echo)
       for (i=1;i<N;i++)
-         st->echo_noise[i] = (.7f*st->echo_noise[i] + .3f* echo[i]);
+         st->echo_noise[i] = (.3f*st->echo_noise[i] + echo[i]);
 
    /* Compute a posteriori SNR */
    for (i=1;i<N;i++)
@@ -755,7 +756,7 @@ int speex_preprocess(SpeexPreprocessState *st, short *x, int *echo)
          if (st->update_prob[i]<.5f || st->ps[i] < st->noise[i])
          {
             if (echo)
-               st->noise[i] = .90f*st->noise[i] + .1f*(st->ps[i]-echo[i]);
+               st->noise[i] = .90f*st->noise[i] + .1f*max(1.0f,st->ps[i]-echo[i]);
             else
                st->noise[i] = .90f*st->noise[i] + .1f*st->ps[i];
          }
@@ -937,7 +938,7 @@ int speex_preprocess(SpeexPreprocessState *st, short *x, int *echo)
    return is_speech;
 }
 
-void speex_preprocess_estimate_update(SpeexPreprocessState *st, short *x, int *echo)
+void speex_preprocess_estimate_update(SpeexPreprocessState *st, spx_int16_t *x, float *echo)
 {
    int i;
    int N = st->ps_size;
@@ -956,7 +957,7 @@ void speex_preprocess_estimate_update(SpeexPreprocessState *st, short *x, int *e
       if (st->update_prob[i]<.5f || st->ps[i] < st->noise[i])
       {
          if (echo)
-            st->noise[i] = .90f*st->noise[i] + .1f*(st->ps[i]-echo[i]);
+            st->noise[i] = .90f*st->noise[i] + .1f*max(1.0f,st->ps[i]-echo[i]);
          else
             st->noise[i] = .90f*st->noise[i] + .1f*st->ps[i];
       }
