@@ -36,6 +36,16 @@
 
 #include "vq.h"
 #include "stack_alloc.h"
+#include "misc.h"
+
+#ifdef _USE_SSE
+#include <xmmintrin.h>
+#elif defined(SHORTCUTS) && (defined(ARM4_ASM) || defined(ARM5E_ASM))
+#include "vq_arm4.h"
+#elif defined(BFIN_ASM)
+#include "vq_bfin.h"
+#endif
+
 
 int scal_quant(spx_word16_t in, const spx_word16_t *boundary, int entries)
 {
@@ -82,51 +92,8 @@ int vq_index(float *in, const float *codebook, int len, int entries)
    return best_index;
 }
 
-#ifdef _USE_SSE
-#include <xmmintrin.h>
-#include "misc.h"
-void vq_nbest(spx_word16_t *_in, const __m128 *codebook, int len, int entries, __m128 *E, int N, int *nbest, spx_word32_t *best_dist, char *stack)
-{
-   int i,j,k,used;
-   float *dist;
-   __m128 *in;
-   __m128 half;
-   used = 0;
-   dist = PUSH(stack, entries, float);
-   half = _mm_set_ps1(.5f);
-   in = PUSH(stack, len, __m128);
-   for (i=0;i<len;i++)
-      in[i] = _mm_set_ps1(_in[i]);
-   for (i=0;i<entries>>2;i++)
-   {
-      __m128 d = _mm_mul_ps(E[i], half);
-      for (j=0;j<len;j++)
-         d = _mm_sub_ps(d, _mm_mul_ps(in[j], *codebook++));
-      _mm_storeu_ps(dist+4*i, d);
-   }
-   for (i=0;i<entries;i++)
-   {
-      if (i<N || dist[i]<best_dist[N-1])
-      {
-         for (k=N-1; (k >= 1) && (k > used || dist[i] < best_dist[k-1]); k--)
-         {
-            best_dist[k]=best_dist[k-1];
-            nbest[k] = nbest[k-1];
-         }
-         best_dist[k]=dist[i];
-         nbest[k]=i;
-         used++;
-      }
-   }
-}
-
-
-#else
-
-#if defined(SHORTCUTS) && (defined(ARM4_ASM) || defined(ARM5E_ASM))
-#include "vq_arm4.h"
-#else
 
+#ifndef OVERRIDE_VQ_NBEST
 /*Finds the indices of the n-best entries in a codebook*/
 void vq_nbest(spx_word16_t *in, const spx_word16_t *codebook, int len, int entries, spx_word32_t *E, int N, int *nbest, spx_word32_t *best_dist, char *stack)
 {
@@ -138,7 +105,7 @@ void vq_nbest(spx_word16_t *in, const spx_word16_t *codebook, int len, int entri
       for (j=0;j<len;j++)
          dist = MAC16_16(dist,in[j],*codebook++);
 #ifdef FIXED_POINT
-      dist=SUB32(SHR(E[i],1),dist);
+      dist=SUB32(SHR32(E[i],1),dist);
 #else
       dist=.5f*E[i]-dist;
 #endif
@@ -157,61 +124,10 @@ void vq_nbest(spx_word16_t *in, const spx_word16_t *codebook, int len, int entri
 }
 #endif
 
-#endif
 
 
 
-#ifdef _USE_SSE
-
-void vq_nbest_sign(spx_word16_t *_in, const __m128 *codebook, int len, int entries, __m128 *E, int N, int *nbest, spx_word32_t *best_dist, char *stack)
-{
-   int i,j,k,used;
-   float *dist;
-   __m128 *in;
-   __m128 half;
-   used = 0;
-   dist = PUSH(stack, entries, float);
-   half = _mm_set_ps1(.5f);
-   in = PUSH(stack, len, __m128);
-   for (i=0;i<len;i++)
-      in[i] = _mm_set_ps1(_in[i]);
-   for (i=0;i<entries>>2;i++)
-   {
-      __m128 d = _mm_setzero_ps();
-      for (j=0;j<len;j++)
-         d = _mm_add_ps(d, _mm_mul_ps(in[j], *codebook++));
-      _mm_storeu_ps(dist+4*i, d);
-   }
-   for (i=0;i<entries;i++)
-   {
-      int sign;
-      if (dist[i]>0)
-      {
-         sign=0;
-         dist[i]=-dist[i];
-      } else
-      {
-         sign=1;
-      }
-      dist[i] += .5f*((float*)E)[i];
-      if (i<N || dist[i]<best_dist[N-1])
-      {
-         for (k=N-1; (k >= 1) && (k > used || dist[i] < best_dist[k-1]); k--)
-         {
-            best_dist[k]=best_dist[k-1];
-            nbest[k] = nbest[k-1];
-         }
-         best_dist[k]=dist[i];
-         nbest[k]=i;
-         used++;
-         if (sign)
-            nbest[k]+=entries;
-      }
-   }
-}
-
-#else
-
+#ifndef OVERRIDE_VQ_NBEST_SIGN
 /*Finds the indices of the n-best entries in a codebook with sign*/
 void vq_nbest_sign(spx_word16_t *in, const spx_word16_t *codebook, int len, int entries, spx_word32_t *E, int N, int *nbest, spx_word32_t *best_dist, char *stack)
 {
@@ -231,7 +147,7 @@ void vq_nbest_sign(spx_word16_t *in, const spx_word16_t *codebook, int len, int
          sign=1;
       }
 #ifdef FIXED_POINT
-      dist = ADD32(dist,SHR(E[i],1));
+      dist = ADD32(dist,SHR32(E[i],1));
 #else
       dist = ADD32(dist,.5f*E[i]);
 #endif
